

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>


#define STATUS  int
#define ERROR   -1
#define OK      0

typedef struct LangTblHeader_s
{
    unsigned long size;
    unsigned long chksum;
    unsigned char version[4];
    unsigned char reserv[4];
} LangTblHeader_t;

#define LANG_TBL_HEADER_LEN         (sizeof(LangTblHeader_t))
#define LANG_TBL_VER_STR_DIGIT_NUM  (8)

#define MAX_FILENAME_LEN            (512)
#define FLASH_BLOCK_SIZE            (64*1024)
#define MAX_SIZE_LANG_TBLFILE       ((1*FLASH_BLOCK_SIZE) - LANG_TBL_HEADER_LEN)

unsigned long calculate_checksum (int action, char *s, int size)
{
    static int c0, c1;
    int checksum, b;
    int i;

    switch (action)
    {
    case 0:
        c0 = c1 = 0;
        break;
    case 1:
        for (i = 0; i < size; i++)
        {
            c0 += s[i] & 0xff;
            c1 += c0;
        }
        break;
    case 2:
        b = (c0 & 65535) + ((c0 >> 16) & 65535);
        c0 = ((b >> 16) + b) & 65535;

        b = (c1 & 65535) + ((c1 >> 16) & 65535);
        c1 = ((b >> 16) + b) & 65535;

        checksum = ((c1 << 16) | c0);

        return checksum;
    }
    return 0;
}


STATUS calculateFileSize(FILE *fd, unsigned long *size)
{
    unsigned long start = 0;
    unsigned long end = 0;

    if (!size)
        return ERROR;
    if (fseek(fd, 0L, SEEK_SET) < 0)
    {
        *size = 0;
        return ERROR;
    }
    if ((start = ftell(fd)) < 0)
    {
        *size = 0;
        return ERROR;
    }
    if (fseek(fd, 0L, SEEK_END) < 0)
    {
        *size = 0;
        return ERROR;
    }
    if ((end = ftell(fd)) < 0)
    {
        *size = 0;
        return ERROR;
    }
    if (fseek(fd, 0L, SEEK_SET) < 0)
    {
        *size = 0;
        return ERROR;
    }
    if ((end - start) < 0)
    {
        *size = 0;
        return ERROR;
    }
    else 
        *size = end - start;

    return OK;
}


STATUS langGetZipLangTblChksum(FILE *zipLangTbl, unsigned long zipLangTblSize,
                        unsigned long *zipLangTblChksum)
{
    unsigned long calcChksum = 0;
    unsigned char *buf = NULL;
    unsigned long bytes_read = 0, read_size = 0;
    int ret = ERROR;

    if ((!zipLangTblChksum) || (zipLangTblSize <= 0) || (!zipLangTbl))
    {
        ret =  ERROR;
        goto end;
    }

    if (!(buf = (unsigned char *)malloc(zipLangTblSize)))
    {
        printf("malloc zipLangTblSize(%lu) fail\n", zipLangTblSize);
        ret = ERROR;
        goto end;
    }

    calculate_checksum (0, NULL, 0);
    if (fseek(zipLangTbl, 0L, SEEK_SET) < 0)
    {
        *zipLangTblChksum = 0;
        ret = ERROR;
        goto end;
    }

    /* Loop to accummulate checksum */
    for (bytes_read=0; bytes_read < zipLangTblSize; bytes_read += read_size)
    {
        if (zipLangTblSize - bytes_read >= FLASH_BLOCK_SIZE)
            read_size = FLASH_BLOCK_SIZE;
        else
            read_size = zipLangTblSize - bytes_read;
        
        if (fread(buf, 1, read_size, zipLangTbl) != read_size)
        {
            *zipLangTblChksum = 0;
            ret = ERROR;
            goto end;
        }
        calculate_checksum (1, buf, read_size);
    }

    if (fseek(zipLangTbl, 0L, SEEK_SET) < 0)
    {
        *zipLangTblChksum = 0;
        ret = ERROR;
        goto end;
    }

    calcChksum = calculate_checksum (2, NULL, 0);
    *zipLangTblChksum = calcChksum;

    ret = OK;

end:
    if (buf)
    {
        free (buf);
        buf = NULL;
    }
    
    return (ret);
}

STATUS langWriteZipLangTblToMTD(char *mtdFileName, FILE *fpZipLangTbl,
            unsigned long zipLangTblSize, LangTblHeader_t *langTblHeader,
            unsigned long langTblHeaderLen)
{
    FILE *fp = NULL;
    unsigned long start = 0;
    unsigned long block_offset = 0, count = 0;
    unsigned long block_size = 0;
    unsigned long zipLangTblWriteSize = 0;
    unsigned int n_written = 0, fail_count = 0;
    
    unsigned char *pZipLangtbl = NULL;
    unsigned char *file_buf = NULL;
    unsigned char *file_ptr = NULL;
    int ret = 0;

    if ((!mtdFileName) || (!fpZipLangTbl) || (!langTblHeader) ||
        (zipLangTblSize <= 0) || (zipLangTblSize >= MAX_SIZE_LANG_TBLFILE) ||
        (langTblHeaderLen <= 0) || (langTblHeaderLen > LANG_TBL_HEADER_LEN))
    {
        ret = ERROR;
        goto end;
    }

    if(!(file_buf = malloc(zipLangTblSize + langTblHeaderLen)))
    {
        printf("malloc zipLangTblSize(%lu) fail\n", zipLangTblSize);
        ret = ERROR;
        goto end;
    }

    memset(file_buf, 0, zipLangTblSize + langTblHeaderLen);
    memcpy(file_buf, langTblHeader, langTblHeaderLen);
    file_ptr = file_buf + langTblHeaderLen;

    if(fread(file_ptr, 1, zipLangTblSize, fpZipLangTbl) != zipLangTblSize)
    {
        printf("fread(fpZipLangTbl) fail\n");
        ret = ERROR;
        goto end;
    }

    /* Open MTD device and get sector size */
    if ((fp = fopen (mtdFileName, "w+")) == NULL)
    {
        printf("open(%s) fail\n", mtdFileName);
        ret = ERROR;
        goto end;
    }

    block_size = FLASH_BLOCK_SIZE;
    pZipLangtbl = file_buf;
    zipLangTblWriteSize = zipLangTblSize + langTblHeaderLen;
    
    /* Write file or URL to MTD device */
    for (start = 0; start < zipLangTblWriteSize;
         start += block_size)
    {
        fail_count = 0;

try_again:
        if (zipLangTblWriteSize - start >= block_size)
            count = block_size;
        else
            count = zipLangTblWriteSize - start;

        if ((n_written = fwrite (pZipLangtbl, 1, count, fp)) != count)
        {
            printf("write pZipLangtbl fail(%d,%lu)\n",
                                        n_written, count);
            fail_count++;
            if (fail_count > 3)
            {
                ret = ERROR;
                goto end;
            }
            else {
                fseek(fp, start, SEEK_SET);
                goto try_again;
            }
        }

        pZipLangtbl += block_size;
        block_offset += count;
    }

    if (block_offset != zipLangTblWriteSize)
    {
        printf ("Writing fail(%lu,%lu)\n",zipLangTblWriteSize, block_offset);
        ret = ERROR;
        goto end;
    }
    printf ("0x%08lX 0x%08lX\n", zipLangTblSize, langTblHeader->chksum);

    ret = OK;
    
end:
    if (fp)
    {
        fclose (fp);
        fp = NULL;
    }

    if(file_buf)
    {
        free(file_buf);
        file_buf = NULL;
    }

    return ret;
}

STATUS langZipLangTbl(char *uzipLangTblPath, unsigned long uzipLangTblSize,
                      char *zipLangTblPath, unsigned long *zipLangTblSize)
{
    FILE *fp = NULL;
    char command[256] = "";
    int ret = ERROR;

    /* Foxconn Perry modified start, 2011/02/22, for bzip2 compression */
    sprintf(command, "rm -f %s;bzip2 -c %s > %s",zipLangTblPath,
                            uzipLangTblPath, zipLangTblPath);
    /* Foxconn Perry modified end, 2011/02/22, for bzip2 compression */
    system(command);

    if((fp = fopen(zipLangTblPath, "r")) == NULL)
    {
        *zipLangTblSize = 0;
        ret = ERROR;
        goto end;
    }

    if(calculateFileSize(fp, zipLangTblSize) != OK)
    {
        *zipLangTblSize = 0;
        ret = ERROR;
        goto end;
    }

    if(*zipLangTblSize >= MAX_SIZE_LANG_TBLFILE)
    {
        printf("The zip file is too large!\n");
        *zipLangTblSize = 0;
        ret = ERROR;
        goto end;
    }

    ret = OK;

end:
    if(fp)
    {
        fclose(fp);
        fp = NULL;
    }
    return ret;
}

STATUS lang_gen_nmrp_st (char *lang_tbl_nmrp, char *ori_st, unsigned long fileSize,
                         char *zip_st, char *version)
{
    FILE *fp_zip = NULL;
    unsigned long zip_lang_tbl_size = 0;
    unsigned long zip_lang_tbl_chksum = 0;
    unsigned int tmp_ver[8] = {0};
    LangTblHeader_t zip_lang_tbl_hdr;
    int ret = ERROR;

    if((langZipLangTbl(ori_st, fileSize,
                   zip_st, &zip_lang_tbl_size)) != OK)
    {
        printf("ZipLangTbl fail!\n");
        ret = ERROR;
        goto end;
    }

    fp_zip = fopen(zip_st, "r");
    if (!fp_zip)
    {
        printf ("%s: no files!\n", zip_st);
        ret = ERROR;
        goto end;
    }
    
    if (langGetZipLangTblChksum(fp_zip, zip_lang_tbl_size,
                                &zip_lang_tbl_chksum) != OK)
    {
        printf("calculate_checksum fail!\n");
        ret = ERROR;
        goto end;
    }

    memset (&zip_lang_tbl_hdr, 0 , sizeof(LangTblHeader_t));
    if(LANG_TBL_VER_STR_DIGIT_NUM != 
        sscanf(version, "%d.%d.%d.%d_%d.%d.%d.%d",
            &tmp_ver[0], &tmp_ver[1], &tmp_ver[2], &tmp_ver[3],
            &tmp_ver[4], &tmp_ver[5], &tmp_ver[6], &tmp_ver[7])
      )
    {
        printf("Invalid version(%s)!\n", version);
        ret = ERROR;
        goto end;
    }

    if ((tmp_ver[0] > 0xFF) || (tmp_ver[1] > 0xFF) ||
        (tmp_ver[2] > 0xFF) || (tmp_ver[3] > 0xFF) ||
        (tmp_ver[4] > 0xFF) || (tmp_ver[5] > 0xFF) ||
        (tmp_ver[6] > 0xFF) || (tmp_ver[7] > 0xFF)
        )
    {
        printf("Invalid version(%s)!\n", version);
        ret = ERROR;
        goto end;
    }

    if ((zip_lang_tbl_size <= 0) || (zip_lang_tbl_size >= MAX_SIZE_LANG_TBLFILE))
    {
        printf("Invalid size(%lu)!\n", zip_lang_tbl_size);
        ret = ERROR;
        goto end;
    }

    zip_lang_tbl_hdr.version[0] = (char)tmp_ver[0];
    zip_lang_tbl_hdr.version[1] = (char)tmp_ver[1];
    zip_lang_tbl_hdr.version[2] = (char)tmp_ver[2];
    zip_lang_tbl_hdr.version[3] = (char)tmp_ver[3];
    zip_lang_tbl_hdr.reserv[0] = (char)tmp_ver[4];
    zip_lang_tbl_hdr.reserv[1] = (char)tmp_ver[5];
    zip_lang_tbl_hdr.reserv[2] = (char)tmp_ver[6];
    zip_lang_tbl_hdr.reserv[3] = (char)tmp_ver[7];
    
    zip_lang_tbl_hdr.size = zip_lang_tbl_size;
    zip_lang_tbl_hdr.chksum = zip_lang_tbl_chksum;

    if (fseek(fp_zip, 0L, SEEK_SET) < 0)
    {
        printf("fseek() failed!\n");
        ret = ERROR;
        goto end;
    }

    if (langWriteZipLangTblToMTD (lang_tbl_nmrp, fp_zip, zip_lang_tbl_size,
            &zip_lang_tbl_hdr, sizeof(LangTblHeader_t)) != OK)
    {
        printf("langWriteZipLangTblToMTD() failed!\n");
        ret = ERROR;
        goto end;
    }

    ret = OK;

end:
    if (fp_zip)
    {
        fclose (fp_zip);
        fp_zip = NULL;
    }

    return (ret);
}

static void usage (void)
{
    printf("usage: nmrp_st_gen input_file output_file version\n");
    exit (0);
}

int main (int argc, char **argv)
{
    FILE *fp = NULL;
    char ori_st_name[MAX_FILENAME_LEN] = "";
    char zip_st_name[MAX_FILENAME_LEN] = "";
    char nmrp_st_name[MAX_FILENAME_LEN] = "";
    char version[MAX_FILENAME_LEN] = "";
    unsigned long ori_st_size = 0;
    char command[512] = "";

    if (argc != 4)
        usage ();

    if ((strlen(argv[1]) >= sizeof(ori_st_name))
        ||(strlen(argv[2]) >= sizeof(nmrp_st_name))
        ||(strlen(argv[3]) >= sizeof(version)))
        usage ();

    strcpy (ori_st_name, argv[1]);
    strcpy (nmrp_st_name, argv[2]);
    strcpy (version, argv[3]);

    /* Foxconn Perry modified start, 2011/02/22, for bzip2 compression */
    sprintf(zip_st_name, "%s.bz2", ori_st_name);
    /* Foxconn Perry modified end, 2011/02/22, for bzip2 compression */
    
    fp = fopen(ori_st_name, "r");
    if (!fp)
    {
        printf("fopen(%s) fail!\n", ori_st_name);
        exit (1);
    }

    if (calculateFileSize(fp, &ori_st_size) != OK)
    {
        fclose(fp);
        exit (2);
    }

    if (lang_gen_nmrp_st(nmrp_st_name, ori_st_name, ori_st_size, zip_st_name, version) != OK)
    {
        exit (3);
    }

    sprintf(command, "rm -rf %s", zip_st_name);
    system(command);

    return 0;
}

